from ..common_definitions import *

#================================================================================#
# region SYSTEM COMMAND DEFINITIONS
#================================================================================#

class SysCommandCodes(Enum):
    """
    Enumeration of the SYS command codes.
    """
    SYS_RESET_DEVICE                            = makeCommandCode(Group.SYS.value, CommandRole.GENERIC.value, CommandType.REQUEST_RESPONSE.value, 1)
    SYS_ENTER_BOOT_MODE                         = makeCommandCode(Group.SYS.value, CommandRole.GENERIC.value, CommandType.REQUEST_RESPONSE.value, 2)
    SYS_ENTER_ISP_MODE                          = makeCommandCode(Group.SYS.value, CommandRole.GENERIC.value, CommandType.REQUEST_RESPONSE.value, 3)
    SYS_SET_I3C_VOLTAGE                         = makeCommandCode(Group.SYS.value, CommandRole.GENERIC.value, CommandType.REQUEST_RESPONSE.value, 4)
    SYS_SET_I2C_SPI_UART_GPIO_VOLTAGE           = makeCommandCode(Group.SYS.value, CommandRole.GENERIC.value, CommandType.REQUEST_RESPONSE.value, 5)
    SYS_SET_1WIRE_VOLTAGE                       = makeCommandCode(Group.SYS.value, CommandRole.GENERIC.value, CommandType.REQUEST_RESPONSE.value, 6)
    SYS_GET_USB_STRING                          = makeCommandCode(Group.SYS.value, CommandRole.GENERIC.value, CommandType.REQUEST_RESPONSE.value, 7)
    SYS_GET_I3C_CONNECTOR_STATUS                = makeCommandCode(Group.SYS.value, CommandRole.GENERIC.value, CommandType.REQUEST_RESPONSE.value, 8)
    SYS_GET_ANALOG_MEASUREMENTS                 = makeCommandCode(Group.SYS.value, CommandRole.GENERIC.value, CommandType.REQUEST_RESPONSE.value, 9)
    SYS_USE_EXTERNAL_I3C_VOLTAGE                = makeCommandCode(Group.SYS.value, CommandRole.GENERIC.value, CommandType.REQUEST_RESPONSE.value, 10)
    SYS_USE_EXTERNAL_I2C_SPI_UART_GPIO_VOLTAGE  = makeCommandCode(Group.SYS.value, CommandRole.GENERIC.value, CommandType.REQUEST_RESPONSE.value, 11)
    SYS_GET_DEVICE_INFO                         = makeCommandCode(Group.SYS.value, CommandRole.GENERIC.value, CommandType.REQUEST_RESPONSE.value, 12)
    SYS_I3C_CONNECTOR_EVENT_NOTIFICATION        = makeCommandCode(Group.SYS.value, CommandRole.GENERIC.value, CommandType.NOTIFICATION.value, 1)

SYS_COMMAND_NAMES = {
    result.value: result.name.replace("_", " ") for result in SysCommandCodes
}

# endregion

#================================================================================#
# region SYSTEM COMMON DEFINITIONS
#================================================================================#

POWER_OFF_VOLTAGE = 0

# endregion

#================================================================================#
# region SYSTEM RESULT DEFINITIONS
#================================================================================#

class SysResultCodes(Enum):
    """
    Enumeration of SYS result codes.
    """
    VOLTAGE_OUT_OF_RANGE                = makeResultCode(Group.SYS.value, 1)
    EXTERNAL_VOLTAGE_DETECTED           = makeResultCode(Group.SYS.value, 2)
    VOLTAGE_ADJUSTMENT_FAILURE          = makeResultCode(Group.SYS.value, 3)
    I3C_PORTS_NOT_POWERED               = makeResultCode(Group.SYS.value, 4)
    BOTH_I3C_PORTS_POWERED              = makeResultCode(Group.SYS.value, 5)
    VOLTAGE_OUT_OF_RANGE_ON_I3C_LV      = makeResultCode(Group.SYS.value, 6)
    VOLTAGE_OUT_OF_RANGE_ON_I3C_HV      = makeResultCode(Group.SYS.value, 7)

SYS_RESULT_NAMES = {
    **COMMON_RESULT_NAMES,
    **{result.value: result.name for result in SysResultCodes}
}

# endregion

#================================================================================#
# region RESET DEVICE
#================================================================================#

# Request ---------------------------------------------------------------------- #

class ResetDeviceRequestFields_t(Structure):
    _pack_ = 1
    _fields_ = [("header", CommandRequestHeader_t)]

ResetDeviceRequestArray_t = c_uint8 * sizeof(ResetDeviceRequestFields_t)

class ResetDeviceRequest_t(BaseCommandRequest_t):
    _fields_ = [("data", ResetDeviceRequestArray_t),
                ("fields", ResetDeviceRequestFields_t)]

# endregion

#================================================================================#
# region ENTER BOOT MODE
#================================================================================#

# Request ---------------------------------------------------------------------- #

class EnterBootModeRequestFields_t(Structure):
    _pack_ = 1
    _fields_ = [("header", CommandRequestHeader_t)]

EnterBootModeRequestArray_t = c_uint8 * sizeof(EnterBootModeRequestFields_t)

class EnterBootModeRequest_t(BaseCommandRequest_t):
    _fields_ = [("data", EnterBootModeRequestArray_t),
                ("fields", EnterBootModeRequestFields_t)]

# endregion

#================================================================================#
# region ENTER ISP MODE
#================================================================================#

# Request ---------------------------------------------------------------------- #

class EnterIspModeRequestFields_t(Structure):
    _pack_ = 1
    _fields_ = [("header", CommandRequestHeader_t)]

EnterIspModeRequestArray_t = c_uint8 * sizeof(EnterIspModeRequestFields_t)

class EnterIspModeRequest_t(BaseCommandRequest_t):
    _fields_ = [("data", EnterIspModeRequestArray_t),
                ("fields", EnterIspModeRequestFields_t)]

# endregion

#================================================================================#
# region GET USB STRING
#================================================================================#

class GetUsbStringSubCommand(Enum):
    """
    Enum that represents the USB Descriptors that can be retrieved by GET USB STRING
    command. The values assigned match the string indexes in the string descriptor.
    """
    MANUFACTURER    = 0x01
    PRODUCT_NAME    = 0X02
    SERIAL_NUMBER   = 0x03
    FW_VERSION      = 0x04
    HW_VERSION      = 0x05
    BL_VERSION      = 0x06

# Request ---------------------------------------------------------------------- #

class GetUsbStringRequestParameters_t(Structure):
    _pack_ = 1
    _fields_ = [("subCommand",c_uint8)]

class GetUsbStringRequestFields_t(Structure):
    _pack_ = 1
    _fields_ = [("header", CommandRequestHeader_t),
                ("parameters", GetUsbStringRequestParameters_t)]

GetUsbStringRequestArray_t = c_uint8 * sizeof(GetUsbStringRequestFields_t)

class GetUsbStringRequest_t(BaseCommandRequest_t):
    _fields_ = [("data", GetUsbStringRequestArray_t),
                ("fields", GetUsbStringRequestFields_t)]

# Response --------------------------------------------------------------------- #

MAXIMUM_USB_STRING_LENGTH = 50

class GetUsbStringResponseParameters_t(Structure):
    _pack_ = 1
    _fields_ = [("payloadLength", c_uint8)]

GetUsbStringResponsePayload_t = c_uint8 * MAXIMUM_USB_STRING_LENGTH

class GetUsbStringResponseFields_t(Structure):
    _pack_ = 1
    _fields_ = [("header", CommandResponseHeader_t),
                ("parameters", GetUsbStringResponseParameters_t),
                ("payload", GetUsbStringResponsePayload_t)]

GetUsbStringResponseArray_t = c_uint8 * sizeof(GetUsbStringResponseFields_t)

class GetUsbStringResponse_t(BaseCommandResponse_t):
    _fields_ = [("data", GetUsbStringResponseArray_t),
                ("fields", GetUsbStringResponseFields_t)]

    def fromBytes(self, data):
        """
        This function set the ctypes Array data from a buffer object
        """
        for i in range(len(data)):
            self.data[i] = data[i]

    def toDictionary(self) -> dict:

        payload_length = self.fields.parameters.payloadLength - 1
        data = str(self.fields.payload,encoding='ascii')[:payload_length]

        return {
            "id": self.fields.header.id,
            "command": SYS_COMMAND_NAMES[self.fields.header.code],
            "result": SYS_RESULT_NAMES[self.fields.header.result],
            "payload_length": payload_length,
            "payload": data
        }


# endregion

#================================================================================#
# region SET I3C VOLTAGE
#================================================================================#

# Constants
MIN_I3C_VOLTAGE_VALUE = 800
MAX_I3C_VOLTAGE_VALUE = 3300

# Request ---------------------------------------------------------------------- #

class SetI3cVoltageRequestParameters_t(Structure):
    _pack_ = 1
    _fields_ = [("voltage_mV" , c_uint16)]

class SetI3cVoltageRequestFields_t(Structure):
    _pack_ = 1
    _fields_ = [("header", CommandRequestHeader_t),
		        ("parameters" , SetI3cVoltageRequestParameters_t)]

SetI3cVoltageRequestArray_t = c_uint8 * sizeof(SetI3cVoltageRequestFields_t)

class SetI3cVoltageRequest_t(BaseCommandRequest_t):
    _fields_ = [("data", SetI3cVoltageRequestArray_t),
                ("fields", SetI3cVoltageRequestFields_t)]

# Response --------------------------------------------------------------------- #

class SetI3cVoltageResponseFields_t(Structure):
    _pack_ = 1
    _fields_ = [("header", CommandResponseHeader_t)]

# Union array
SetI3cVoltageResponseArray_t = c_uint8 * sizeof(SetI3cVoltageResponseFields_t)

class SetI3cVoltageResponse_t(BaseCommandResponse_t):

    _fields_ = [("data", SetI3cVoltageResponseArray_t),
                ("fields", SetI3cVoltageResponseFields_t)]

    def fromBytes(self, data):
        """
        This function set the ctypes Array data from a buffer object
        """
        self.data = SetI3cVoltageResponseArray_t.from_buffer_copy(data)

    def toDictionary(self) -> dict:
        return {
            "id": self.fields.header.id,
            "command": SYS_COMMAND_NAMES[self.fields.header.code],
            "result": SYS_RESULT_NAMES[self.fields.header.result]
        }

# endregion

#================================================================================#
# region SET I2C/SPI/UART/GPIO VOLTAGE
#================================================================================#

# Constants
MIN_I2C_SPI_UART_GPIO_VOLTAGE_VALUE = 1200
MAX_I2C_SPI_UART_GPIO_VOLTAGE_VALUE = 3300

# Request ---------------------------------------------------------------------- #

class SetI2cSpiUartGpioVoltageRequestParameters_t(Structure):
    _pack_ = 1
    _fields_ = [("voltage_mV", c_uint16)]

class SetI2cSpiUartGpioVoltageRequestFields_t(Structure):
    _pack_ = 1
    _fields_ = [("header", CommandRequestHeader_t),
		        ("parameters" , SetI2cSpiUartGpioVoltageRequestParameters_t),]

SetI2cSpiUartGpioVoltageRequestArray_t = c_uint8 * sizeof(SetI2cSpiUartGpioVoltageRequestFields_t)

class SetI2cSpiUartGpioVoltageRequest_t(BaseCommandRequest_t):
    _fields_ = [("data", SetI2cSpiUartGpioVoltageRequestArray_t),
                ("fields", SetI2cSpiUartGpioVoltageRequestFields_t)]

# Response --------------------------------------------------------------------- #

class SetI2cSpiUartGpioVoltageResponseFields_t(Structure):
    _pack_ = 1
    _fields_ = [("header", CommandResponseHeader_t)]

SetI2cSpiUartGpioVoltageResponseArray_t = c_uint8 * sizeof(SetI2cSpiUartGpioVoltageResponseFields_t)

# Union command
class SetI2cSpiUartGpioVoltageResponse_t(BaseCommandResponse_t):
    _fields_ = [("data", SetI2cSpiUartGpioVoltageResponseArray_t),
                ("fields", SetI2cSpiUartGpioVoltageResponseFields_t)]

    def fromBytes(self, data):
        """
        This function set the ctypes Array data from a buffer object
        """
        self.data = SetI2cSpiUartGpioVoltageResponseArray_t.from_buffer_copy(data)

    def toDictionary(self) -> dict:
        return {
            "id": self.fields.header.id,
            "command": SYS_COMMAND_NAMES[self.fields.header.code],
            "result": SYS_RESULT_NAMES[self.fields.header.result]
        }

# endregion

#================================================================================#
# region USE EXTERNAL I3C VOLTAGE
#================================================================================#

# Request ---------------------------------------------------------------------- #

class UseExtI3cVoltageRequestFields_t(Structure):
    _pack_ = 1
    _fields_ = [("header", CommandRequestHeader_t)]

UseExtI3cVoltageRequestArray_t = c_uint8 * sizeof(UseExtI3cVoltageRequestFields_t)

class UseExtI3cVoltageRequest_t(BaseCommandRequest_t):
    _fields_ = [("data", UseExtI3cVoltageRequestArray_t),
                ("fields", UseExtI3cVoltageRequestFields_t)]

# Response --------------------------------------------------------------------- #

class UseExtI3cVoltageResponseParameters_t(Structure):
    _pack_ = 1
    _fields_ = [("externalHighVoltage_mV", c_uint16),
                ("externalLowVoltage_mV", c_uint16)]

class UseExtI3cVoltageResponseFields_t(Structure):
    _pack_ = 1
    _fields_ = [("header", CommandResponseHeader_t),
                ("parameters", UseExtI3cVoltageResponseParameters_t)]

UseExtI3cVoltageResponseArray_t = c_uint8 * sizeof(UseExtI3cVoltageResponseFields_t)

class UseExtI3cVoltageResponse_t(BaseCommandResponse_t):
    _fields_ = [("data", UseExtI3cVoltageResponseArray_t),
                ("fields", UseExtI3cVoltageResponseFields_t)]

    def fromBytes(self, data):
        """
        This function set the ctypes Array data from a buffer object
        """
        self.data = UseExtI3cVoltageResponseArray_t.from_buffer_copy(data)

    def toDictionary(self) -> dict:
        return {
            "id": self.fields.header.id,
            "command": SYS_COMMAND_NAMES[self.fields.header.code],
            "result": SYS_RESULT_NAMES[self.fields.header.result],
            "i3c_low_voltage_mV": self.fields.parameters.externalLowVoltage_mV,
            "i3c_high_voltage_mV": self.fields.parameters.externalHighVoltage_mV
        }

# endregion

#================================================================================#
# region USE EXTERNAL I2C/SPI/UART/GPIO VOLTAGE
#================================================================================#

# Request ---------------------------------------------------------------------- #

class UseExternalI2cSpiUartGpioVoltageRequestFields_t(Structure):
    _pack_ = 1
    _fields_ = [("header", CommandRequestHeader_t)]

UseExternalI2cSpiUartGpioVoltageRequestArray_t = c_uint8 * sizeof(UseExternalI2cSpiUartGpioVoltageRequestFields_t)

class UseExternalI2cSpiUartGpioVoltageRequest_t(BaseCommandRequest_t):
    _fields_ = [("data", UseExternalI2cSpiUartGpioVoltageRequestArray_t),
                ("fields", UseExternalI2cSpiUartGpioVoltageRequestFields_t)]

# Response --------------------------------------------------------------------- #

class UseExternalI2cSpiUartGpioVoltageResponseParameters_t(Structure):
    _pack_ = 1
    _fields_ = [("voltage_mV", c_uint16)]

class UseExternalI2cSpiUartGpioVoltageResponseFields_t(Structure):
    _pack_ = 1
    _fields_ = [("header", CommandResponseHeader_t),
                ("parameters", UseExternalI2cSpiUartGpioVoltageResponseParameters_t)]

UseExternalI2cSpiUartGpioVoltageResponseArray_t = c_uint8 * sizeof(UseExternalI2cSpiUartGpioVoltageResponseFields_t)

class UseExternalI2cSpiUartGpioVoltageResponse_t(BaseCommandResponse_t):
    _fields_ = [("data", UseExternalI2cSpiUartGpioVoltageResponseArray_t),
                ("fields", UseExternalI2cSpiUartGpioVoltageResponseFields_t)]

    def fromBytes(self, data):
        """
        This function set the ctypes Array data from a buffer object
        """
        self.data = UseExternalI2cSpiUartGpioVoltageResponseArray_t.from_buffer_copy(data)

    def toDictionary(self) -> dict:
        return {
            "id": self.fields.header.id,
            "command": SYS_COMMAND_NAMES[self.fields.header.code],
            "result": SYS_RESULT_NAMES[self.fields.header.result],
            "voltage_mV": self.fields.parameters.voltage_mV
        }

# endregion

#================================================================================#
# region GET ANALOG MEASUREMENTS
#================================================================================#

# Request ---------------------------------------------------------------------- #

class GetAnalogMeasurementsRequestFields_t(Structure):
    _pack_ = 1
    _fields_ = [("header", CommandRequestHeader_t)]

GetAnalogMeasurementsRequestArray_t = c_uint8 * sizeof(GetAnalogMeasurementsRequestFields_t)

class GetAnalogMeasurementsRequest_t(BaseCommandRequest_t):
    _fields_ = [("data", GetAnalogMeasurementsRequestArray_t),
                ("fields", GetAnalogMeasurementsRequestFields_t)]

# Response --------------------------------------------------------------------- #

class GetAnalogMeasurementsResponseParameters_t(Structure):
    _pack_ = 1
    _fields_ = [("vmeasVtargI2cSpiUartGpio_mV", c_uint16),
                ("vmeasVtargI3cHV_mV", c_uint16),
                ("vmeasVtargI3cLV_mV", c_uint16),
                ("vmeasVcca_mV", c_uint16),
                ("vmeasVccaI3c_mV", c_uint16),
                ("vmeasVddio2_mV", c_uint16)]

class GetAnalogMeasurementsResponseFields_t(Structure):
    _pack_ = 1
    _fields_ = [("header", CommandResponseHeader_t),
                ("parameters", GetAnalogMeasurementsResponseParameters_t)]

GetAnalogMeasurementsResponseArray_t = c_uint8 * sizeof(GetAnalogMeasurementsResponseFields_t)

class GetAnalogMeasurementsResponse_t(BaseCommandResponse_t):

    _fields_ = [("data", GetAnalogMeasurementsResponseArray_t),
                ("fields", GetAnalogMeasurementsResponseFields_t)]

    def fromBytes(self, data):
        """
        This function set the ctypes Array data from a buffer object
        """
        self.data = GetAnalogMeasurementsResponseArray_t.from_buffer_copy(data)

    def toDictionary(self) -> dict:
        return {
            "id": self.fields.header.id,
            "command": SYS_COMMAND_NAMES[self.fields.header.code],
            "result": SYS_RESULT_NAMES[self.fields.header.result],
            "i2c_spi_uart_gpio_vtarg": {
                "internal_mV": self.fields.parameters.vmeasVcca_mV,
                "external_mV": self.fields.parameters.vmeasVtargI2cSpiUartGpio_mV,
            },
            "i3c_low_voltage_vtarg": {
                "internal_mV": self.fields.parameters.vmeasVccaI3c_mV,
                "external_mV": self.fields.parameters.vmeasVtargI3cLV_mV,
            },
            "i3c_high_voltage_vtarg": {
                "internal_mV": self.fields.parameters.vmeasVddio2_mV,
                "external_mV": self.fields.parameters.vmeasVtargI3cHV_mV,
            }
        }

# endregion

#================================================================================#
# region GET I3C CONNECTORS STATUS
#================================================================================#

class I3cConnectorPort_t(Enum):
    I3C_LOW_VOLTAGE_PORT   = 0x00
    I3C_HIGH_VOLTAGE_PORT  = 0x01

class I3cConnectorEvent_t(Enum):
    I3C_CONNECTOR_PLUGGED       = 0x00
    I3C_CONNECTOR_UNPLUGGED     = 0x01

class I3cConnectorType_t(Enum):
    CONNECTOR_IDENTIFICATION_NOT_SUPPORTED  = 0x00
    I3C_HARNESS                             = 0x01
    QWIIC_ADAPTER                           = 0x02
    SENSEPEEK_PROBES                        = 0x03
    NO_CONNECTOR                            = 0x04
    ERROR_IDENTIFYING_CONNECTOR             = 0x05

# Request ---------------------------------------------------------------------- #

# Union structure
class GetI3cConnectorsStatusRequestFields_t(Structure):
    _pack_ = 1
    _fields_ = [("header", CommandRequestHeader_t)]

GetI3cConnectorsStatusRequestArray_t = c_uint8 * sizeof(GetI3cConnectorsStatusRequestFields_t)

class GetI3cConnectorsStatusRequest_t(BaseCommandRequest_t):
    _fields_ = [("data", GetI3cConnectorsStatusRequestArray_t),
                ("fields", GetI3cConnectorsStatusRequestFields_t)]

# Response --------------------------------------------------------------------- #

class GetI3cConnectorsStatusResponseParameters_t(Structure):
    _pack_ = 1
    _fields_ = [("lvConnectorStatus", c_uint8),     # The connectorStatus field gives support for the Rev. B since It doesn't have connector identification
                ("lvConnectorType", c_uint8),
                ("hvConnectorStatus", c_uint8),
                ("hvConnectorType", c_uint8)]

class GetI3cConnectorsStatusResponseFields_t(Structure):
    _pack_ = 1
    _fields_ = [("header", CommandResponseHeader_t),
                ("parameters", GetI3cConnectorsStatusResponseParameters_t)]

GetI3cConnectorsStatusResponseArray_t = c_uint8 * sizeof(GetI3cConnectorsStatusResponseFields_t)

class GetI3cConnectorsStatusResponse_t(BaseCommandResponse_t):
    _fields_ = [("data", GetI3cConnectorsStatusResponseArray_t),
                ("fields", GetI3cConnectorsStatusResponseFields_t)]

    def fromBytes(self, data):
        """
        This function set the ctypes Array data from a buffer object
        """
        self.data = GetI3cConnectorsStatusResponseArray_t.from_buffer_copy(data)

    def toDictionary(self) -> dict:
        return {
            "id": self.fields.header.id,
            "command": SYS_COMMAND_NAMES[self.fields.header.code],
            "result": SYS_RESULT_NAMES[self.fields.header.result],
            "i3c_low_voltage_port": {
                "state": I3cConnectorEvent_t(self.fields.parameters.lvConnectorStatus).name,
                "connector_type": I3cConnectorType_t(self.fields.parameters.lvConnectorType).name
            },
            "i3c_high_voltage_port": {
                "state": I3cConnectorEvent_t(self.fields.parameters.hvConnectorStatus).name,
                "connector_type": I3cConnectorType_t(self.fields.parameters.hvConnectorType).name
            }
        }

# endregion

#================================================================================#
# region GET DEVICE INFO
#================================================================================#

# Request ---------------------------------------------------------------------- #
class GetDeviceInfoRequestFields_t(Structure):
    _pack_ = 1
    _fields_ = [("header", CommandRequestHeader_t)]

GetDeviceInfoRequestArray_t = c_uint8 * sizeof(GetDeviceInfoRequestFields_t)

class GetDeviceInfoRequest_t(BaseCommandRequest_t):
    _fields_ = [("data", GetDeviceInfoRequestArray_t),
                ("fields", GetDeviceInfoRequestFields_t)]
    
# Response --------------------------------------------------------------------- #

class DeviceManufacturer_t(Enum):
    Binho_LLC = 0x00

class ProductName_t(Enum):
    Supernova = 0x00
    Pulsar = 0x01

MANUFACTURER_INDEX  = 0
PRODUCT_NAME_INDEX  = 1
HW_VERSION_INDEX    = 2
SERIAL_NUMBER_INDEX = 3
SERIAL_NUMBER_SIZE  = 32
FW_VERSION_INDEX    = SERIAL_NUMBER_INDEX + SERIAL_NUMBER_SIZE

GET_DEVICE_INFO_PAYLOAD_LENGTH = 73

class GetDeviceInfoResponseParameters_t(Structure):
    _pack_ = 1
    _fields_ = [("firmwareVersionLength", c_uint8),
                ("payloadLength", c_uint8)]

GetDeviceInfoResponsePayload_t = c_uint8 * GET_DEVICE_INFO_PAYLOAD_LENGTH

class GetDeviceInfoResponseFields_t(Structure):
    _pack_ = 1
    _fields_ = [("header", CommandResponseHeader_t),
                ("parameters", GetDeviceInfoResponseParameters_t),
                ("payload", GetDeviceInfoResponsePayload_t)]
    
GetDeviceInfoResponseArray_t = c_uint8 * sizeof(GetDeviceInfoResponseFields_t)

class GetDeviceInfoResponse_t(BaseCommandResponse_t):
    _fields_ = [("data", GetDeviceInfoResponseArray_t),
                ("fields", GetDeviceInfoResponseFields_t)]

    def fromBytes(self, data):
        """
        This function set the ctypes Array data from a buffer object
        """
        for i in range(len(data)):
            self.data[i] = data[i]

    def toDictionary(self) -> dict:

        return {
            "id": self.fields.header.id,
            "command": SYS_COMMAND_NAMES[self.fields.header.code],
            "result": SYS_RESULT_NAMES[self.fields.header.result],
            "manufacturer": DeviceManufacturer_t(self.fields.payload[MANUFACTURER_INDEX]).name.replace("_", " "),
            "product_name": ProductName_t(self.fields.payload[PRODUCT_NAME_INDEX]).name.replace("_", " "),
            "serial_number": bytes(self.fields.payload[SERIAL_NUMBER_INDEX:SERIAL_NUMBER_INDEX + SERIAL_NUMBER_SIZE]).decode('ascii'),
            "hardware_version": bytes([self.fields.payload[HW_VERSION_INDEX]]).decode('ascii'),
            "firmware_version": bytes(self.fields.payload[FW_VERSION_INDEX:FW_VERSION_INDEX + self.fields.parameters.firmwareVersionLength]).decode('ascii'),
            "capabilities": {"supported_groups": [Group(group).name for group in self.fields.payload[FW_VERSION_INDEX + self.fields.parameters.firmwareVersionLength:self.fields.parameters.payloadLength]]}
        }

# endregion

#================================================================================#
# region I3C CONNECTOR NOTIFICATION
#================================================================================#

class I3cConnectorNotificationParametersFields_t(Structure):
    _pack_ = 1
    _fields_ = [("port", c_uint8),
                ("event", c_uint8),
                ("connectorType", c_uint8)]

class I3cConnectorNotificationFields_t(Structure):
    _pack_ = 1
    _fields_ = [("header", NotificationHeader_t),
                ("parameters", I3cConnectorNotificationParametersFields_t)]

I3cConnectorNotificationArray_t = c_uint8 * sizeof(I3cConnectorNotificationFields_t)

class I3cConnectorNotification_t(Union):
    _fields_ = [("data", I3cConnectorNotificationArray_t),
                ("fields", I3cConnectorNotificationFields_t)]

    def toDictionary(self) -> dict:
        return {
            "id": self.fields.header.id,
            "command": SYS_COMMAND_NAMES[self.fields.header.code],
            "result": SYS_RESULT_NAMES[self.fields.header.result],
            "event": I3cConnectorEvent_t(self.fields.parameters.event).name,
            "port": I3cConnectorPort_t(self.fields.parameters.port).name,
            "connector_type": I3cConnectorType_t(self.fields.parameters.connectorType).name
        }

    def __str__(self) -> str:
        return str(self.toDictionary())

# endregion


from .definitions import *

def gpioConfigurePinSerializer(id: c_uint16, pinNumber: GpioPinNumber, functionality: GpioFunctionality) -> tuple[bytes, GpioConfigurePinResponse_t]:
    """
    This function creates the associated command for configuring a GPIO pin and sends it to the target device.

    Arguments
    ---------
    id : c_uint16
        ID that identifies the transaction.
    pinNumber : GpioPinNumber
        The GPIO pin number to be configured. Must be one of the options provided by the GpioPinNumber enum.
    functionality : GpioFunctionality
        The desired functionality of the GPIO pin, chosen from the options available in the GpioFunctionality enum.

    Returns
    -------
    tuple
        Command serialization to be sent to the device and the response instance.

    """
    # Create command instance
    command = GpioConfigurePinRequest_t()

    # Fill command fields.
    command.fields.header.id = id
    command.fields.header.code = GpioCommandCodes.GPIO_CONFIGURE_PIN.value
    command.fields.parameters.pinNumber = pinNumber.value
    command.fields.parameters.functionality = functionality.value

    return command.toBytes(), GpioConfigurePinResponse_t()

def gpioDigitalWriteSerializer(id: c_uint16, pinNumber: GpioPinNumber, logicLevel: GpioLogicLevel) -> tuple[bytes, GpioDigitalWriteResponse_t]:
    """
    This function creates the associated command for setting the digital logic level of a GPIO pin and sends it to the target device.

    Arguments
    ---------
    id : c_uint16
        ID that identifies the transaction.
    pinNumber : GpioPinNumber
        The GPIO pin number where the digital logic level will be set. Must be one of the options provided by the GpioPinNumber enum.
    logicLevel : GpioLogicLevel
        The desired logic level (HIGH or LOW) to be set on the specified GPIO pin. Selected from the options available in the GpioLogicLevel enum.

    Returns
    -------
    tuple
        Command serialization to be sent to the device and the response instance.

    """
    # Create command instance
    command = GpioDigitalWriteRequest_t()

    # Fill command fields.
    command.fields.header.id = id
    command.fields.header.code = GpioCommandCodes.GPIO_DIGITAL_WRITE.value
    command.fields.parameters.pinNumber = pinNumber.value
    command.fields.parameters.logicLevel = logicLevel.value

    return command.toBytes(), GpioDigitalWriteResponse_t()

def gpioDigitalReadSerializer(id: c_uint16, pinNumber: GpioPinNumber) -> tuple[bytes, GpioDigitalReadResponse_t]:
    """
    This function creates the associated command for reading the digital logic level of a GPIO pin and sends it to the target device.

    Arguments
    ---------
    id : c_uint16
        ID that identifies the transaction.
    pinNumber : GpioPinNumber
        The GPIO pin number where the digital logic level will be set. Must be one of the options provided by the GpioPinNumber enum.

    Returns
    -------
    tuple
        Command serialization to be sent to the device and the response instance.

    """
    # Create command instance
    command = GpioDigitalReadRequest_t()

    # Fill command fields.
    command.fields.header.id = id
    command.fields.header.code = GpioCommandCodes.GPIO_DIGITAL_READ.value
    command.fields.parameters.pinNumber = pinNumber.value

    return command.toBytes(), GpioDigitalReadResponse_t()

def gpioSetInterruptSerializer(id: c_uint16, pinNumber: GpioPinNumber, trigger: GpioTriggerType) -> tuple[bytes, GpioSetInterruptResponse_t]:
    """
    This function creates the associated command for setting an interruption of a GPIO pin and sends it to the target device.

    Arguments
    ---------
    id : c_uint16
        ID that identifies the transaction.
    pinNumber : GpioPinNumber
        The GPIO pin number where the digital logic level will be set. Must be one of the options provided by the GpioPinNumber enum.
    trigger : GpioTriggerType
        The trigger type used for the interruption. Must be one of the options provided by the GpioTriggerType enum.

    Returns
    -------
    tuple
        Command serialization to be sent to the device and the response instance.

    """
    # Create command instance
    command = GpioSetInterruptRequest_t()

    # Fill command fields.
    command.fields.header.id = id
    command.fields.header.code = GpioCommandCodes.GPIO_SET_INTERRUPT.value
    command.fields.parameters.pinNumber = pinNumber.value
    command.fields.parameters.trigger = trigger.value

    return command.toBytes(), GpioSetInterruptResponse_t()

def gpioDisableInterruptSerializer(id: c_uint16, pinNumber: GpioPinNumber) -> tuple[bytes, GpioDisableInterruptResponse_t]:
    """
    This function creates the associated command for disabling an interruption of a GPIO pin and sends it to the target device.

    Arguments
    ---------
    id : c_uint16
        ID that identifies the transaction.
    pinNumber : GpioPinNumber
        The GPIO pin number where the digital logic level will be set. Must be one of the options provided by the GpioPinNumber enum.

    Returns
    -------
    tuple
        Command serialization to be sent to the device and the response instance.

    """
    # Create command instance
    command = GpioDisableInterruptRequest_t()

    # Fill command fields.
    command.fields.header.id = id
    command.fields.header.code = GpioCommandCodes.GPIO_DISABLE_INTERRUPT.value
    command.fields.parameters.pinNumber = pinNumber.value

    return command.toBytes(), GpioDisableInterruptResponse_t()


from ctypes import *

#================================================================================#
# region USB Packet Header.
#================================================================================#

class UsbTransferPacketHeaderFields_t(Structure):
    """
    USB Transfer Packet Header Fields definition.

    This structures represents the different fields that compose the header of the
    USB packets. It allows to seamlessly access the different fields in order to
    write to them or read from them.
    """
    _pack_ = 1
    _fields_ = [("sot", c_uint16, 1),
                ("eot", c_uint16, 1),
                ("payloadLength", c_uint16, 10),
                ("reserved", c_uint16, 4)]

UsbTransferUsbTransferPacketHeaderArray_t = c_uint8 * sizeof(UsbTransferPacketHeaderFields_t)

class UsbTransferPacketHeader_t(Union):
    """
    USB Transfer Packet Header union definition.

    This union represents the header of the USB packets transferred between the
    USB Host and the USB Host Adapter. This unions allows serialization and
    deserialization of the information by combining the UsbTransferPacketHeaderFields_t
    structure and the UsbTransferUsbTransferPacketHeaderArray_t array.
    """
    _fields_ = [("data", UsbTransferUsbTransferPacketHeaderArray_t),
               ("fields", UsbTransferPacketHeaderFields_t)]

    def toBytes(self) -> bytes:
        return bytes(self.data)

#endregion

#================================================================================#
# region Constants definition
#================================================================================#

USB_FS_INTERRUPT_ENDPOINT_SIZE          = 64
USB_FS_TRANSFER_PACKET_PAYLOAD_LENGTH   = USB_FS_INTERRUPT_ENDPOINT_SIZE - sizeof(UsbTransferPacketHeaderFields_t)

USB_HS_INTERRUPT_ENDPOINT_SIZE          = 1024
USB_HS_TRANSFER_PACKET_PAYLOAD_LENGTH   = USB_HS_INTERRUPT_ENDPOINT_SIZE - sizeof(UsbTransferPacketHeaderFields_t)

#endregion

#================================================================================#
# region USB TRANSFER PROTOCOL class
#================================================================================#

class UsbTransferProtocol:
    """
    Usb Transfer Protocol class in charge of managing the USB Packets transfer.

    This class is intended to control the management of USB packets transferred
    between the USB Host and the USB Host Adapter devices. This class is in total
    aware of the structure of the USB packets and for this reason is able to
    extract the information from incoming packets in order to build the whole input transfer,
    as well as create all the output packets to send data to the USB Host Adapter.

    Attributes
    ----------
    inputTransfer: bytearray
        This bytearray object allows to accumulate the payload of incoming packets in order
        to rebuild the whole input transfer.
    outputTransferPackets: list
        This list object is intended to accumulate all the USB Packets whose payload contains
        a chunk of the complete output transfer.
    endpointSize: int
        Size of the USB interrupt transfer endpoint.

    Methods
    -------
    createUsbOutputTransferPackets(data: bytes)
        Create as many USB packets as needed to transfer data to the USB device.
    receiveUsbInputTransferPackets(data: bytes)
        Receives the input transfer packets sent by the device and creates the whole
        input transfer.

    """

    def __init__(self, endpointSize):

        self.inputTransfer = bytearray()
        self.outputTransferPackets = []
        self.endpointSize = endpointSize

    def createUsbOutputTransferPackets(self, data: bytes) -> list:
        """
        Create as many USB packets as needed to transfer data to the USB device.

        This function receives the whole output transfer data to be sent, create data chunks and build
        as many USB packets as required to transfer the whole data to the device.
        Return a list containing all the packets to be sent to the device.

        Parameters
        ----------
        data: bytes
            A bytes object containing the data to be transferred to the device.

        Returns
        -------
        list
            A list containing all the USB packets.

        """
        self.outputTransferPackets.clear()

        if self.endpointSize == USB_FS_INTERRUPT_ENDPOINT_SIZE:
            packetPayloadLength = USB_FS_TRANSFER_PACKET_PAYLOAD_LENGTH
        elif self.endpointSize == USB_HS_INTERRUPT_ENDPOINT_SIZE:
            packetPayloadLength = USB_HS_TRANSFER_PACKET_PAYLOAD_LENGTH

        nPackets    = len(data) // packetPayloadLength
        nModPackets = len(data) % packetPayloadLength
        if nModPackets != 0: nPackets += 1

        for i in range(nPackets):

            header = UsbTransferPacketHeader_t()

            if i == 0:
                header.fields.sot = 1
            if i == (nPackets - 1):
                header.fields.eot = 1

            startIndex = i * packetPayloadLength
            remainingLength = len(data) - startIndex
            header.fields.payloadLength = min(remainingLength, packetPayloadLength);
            endIndex = startIndex + header.fields.payloadLength

            packet = header.toBytes() + data[startIndex:endIndex]
            self.outputTransferPackets.append(packet)

        return self.outputTransferPackets

    def receiveUsbInputTransferPackets(self, data: bytes) -> bytearray:
        """
        Receives the input transfer packets sent by the device and creates the whole
        input transfer.

        Receives a bytes object containing the serialization of an input transfer packet
        sent by the device. Accumulate the payload of the input packets until the last one is
        received. Returns a bytearray object containing the complete transfer data when the last
        USB packet arrives. Otherwise, returns None.

        Parameters
        ----------
        data: bytes
            A bytes object containing the data sent by the USB device.

        Returns
        -------
        None
            Returns None while an interrupt is still ongoing and the last packet has not been
            received yet.
        bytearray:
            Returns a bytearray object containing all the input transfer sent by the USB device.

        """

        headerLength = sizeof(UsbTransferPacketHeader_t)
        header = UsbTransferPacketHeader_t.from_buffer_copy(data[:headerLength])

        if header.fields.sot:
            self.inputTransfer.clear()


        self.inputTransfer += data[headerLength : headerLength + header.fields.payloadLength]

        if header.fields.eot:
            return self.inputTransfer
        else:
            return None

#endregion


from .definitions import *

#================================================================================#
# region I3C CONTROLLER INIT SERIALIZER
#================================================================================#

def i3cControllerInitSerializer(id: c_uint16, pushPullFrequency: I3cPushPullTransferRate, i3cOpenDrainFrequency: I3cOpenDrainTransferRate, i2cOpenDrainFrequency: I2cTransferRate, driveStrength: I3cDriveStrength) -> tuple[bytes,I3cControllerInitResponse_t]:
    """
    This functions create and serializes an I3C CONTROLLER INIT request and return the corresponding response instance.

    Arguments
    ---------
    id : c_uint16
        ID that identifies the transaction.
    pushPullFrequency: I3cPushPullTransferRate
        I3C Push-Pull frequency
    i3cOpenDrainFrequency: I3cOpenDrainTransferRate
        I3C Open Drain frequency
    i2cOpenDrainFrequency: I2cTransferRate
        I2C Open Drain frequency
    driveStrength: I3cDriveStrength
        Drive strength of the I3C controller.

    Returns
    -------
    tuple
        Command serialization to be sent to the device and the response instance.

    """
    command = I3cControllerInitRequest_t()

    command.fields.header.id = id
    command.fields.header.code = I3cCommandCodes.I3C_CONTROLLER_INIT.value

    command.fields.parameters.pushPullFrequency = pushPullFrequency.value
    command.fields.parameters.i3cOpenDrainFrequency = i3cOpenDrainFrequency.value
    command.fields.parameters.i2cOpenDrainFrequency = i2cOpenDrainFrequency.value
    command.fields.parameters.driveStrength = driveStrength.value

    return command.toBytes(), I3cControllerInitResponse_t()

#endregion

#================================================================================#
# region I3C CONTROLLER SET PARAMETERS SERIALIZER
#================================================================================#

def i3cControllerSetParametersSerializer(id: c_uint16, pushPullFrequency: I3cPushPullTransferRate, i3cOpenDrainFrequency: I3cOpenDrainTransferRate, i2cOpenDrainFrequency: I2cTransferRate, driveStrength: I3cDriveStrength) -> tuple[bytes,I3cControllerSetParametersResponse_t]:
    """
    This functions create and serializes an I3C CONTROLLER SET PARAMETERS request and return the corresponding response instance.

    Arguments
    ---------
    id : c_uint16
        ID that identifies the transaction.
    pushPullFrequency: I3cPushPullTransferRate
        I3C Push-Pull frequency
    i3cOpenDrainFrequency: I3cOpenDrainTransferRate
        I3C Open Drain frequency
    i2cOpenDrainFrequency: I2cTransferRate
        I2C Open Drain frequency
    driveStrength: I3cDriveStrength
        Drive strength of the I3C controller.

    Returns
    -------
    tuple
        Command serialization to be sent to the device and the response instance.

    """
    command = I3cControllerSetParametersRequest_t()

    command.fields.header.id = id
    command.fields.header.code = I3cCommandCodes.I3C_CONTROLLER_SET_PARAMETERS.value

    command.fields.parameters.pushPullFrequency = pushPullFrequency.value
    command.fields.parameters.i3cOpenDrainFrequency = i3cOpenDrainFrequency.value
    command.fields.parameters.i2cOpenDrainFrequency = i2cOpenDrainFrequency.value
    command.fields.parameters.driveStrength = driveStrength.value

    return command.toBytes(), I3cControllerSetParametersResponse_t()

#endregion

#================================================================================#
# region I3C CONTROLLER INIT BUS SERIALIZER
#================================================================================#

def i3cControllerInitBusSerializer(id: c_uint16, targetDevicesTable: dict = None) -> tuple[bytes,I3cControllerInitBusResponse_t]:
    """
    This functions create and serializes an I3C CONTROLLER INIT BUS request and return the corresponding response instance.

    Arguments
    ---------
    id : c_uint16
        ID that identifies the transaction.
    targetDevicesTable : dict
        Python dict that contains the Target Device Table information.

    Returns
    -------
    tuple
        Command serialization to be sent to the device and the response instance.

    """
    command = I3cControllerInitBusRequest_t()

    # Set endpoint ID and command opcode.
    command.fields.header.id = id
    command.fields.header.code = I3cCommandCodes.I3C_CONTROLLER_INIT_BUS.value
    command.fields.parameters.numberOfTargets = 0

    if targetDevicesTable is not None:

        command.fields.parameters.numberOfTargets = len(targetDevicesTable)

        for i in range(command.fields.parameters.numberOfTargets):

            target = targetDevicesTable[i]

            tableEntry = I3cTargetDeviceEntry_t()

            # Addresses
            tableEntry.staticAddress = target["staticAddress"]
            tableEntry.dynamicAddress = target["dynamicAddress"]

            # Provisioned ID
            pid = target["pid"]
            tableEntry.pid.data[0:sizeof(I3cPidBytes_t)] = pid[0:sizeof(I3cPidBytes_t)]

            # BCR
            tableEntry.bcr.byte = target["bcr"]

            # DCR
            tableEntry.dcr = target["dcr"]

            # Lengths
            tableEntry.mwl = 0
            tableEntry.mrl = 0
            tableEntry.maxIbiPayloadLength = 0

            # Configuration
            configuration = I3cTargetConfiguration_t()
            configuration.fields.targetType = target["configuration"]["targetType"].value
            configuration.fields.acceptIbiRequest = target["configuration"]["IBIRequest"].value
            configuration.fields.acceptControllerRoleRequest = target["configuration"]["CRRequest"].value
            configuration.fields.daaUseSETDASA = target["configuration"]["daaUseSETDASA"].value
            configuration.fields.daaUseSETAASA = target["configuration"]["daaUseSETAASA"].value
            configuration.fields.daaUseENTDAA = target["configuration"]["daaUseENTDAA"].value
            configuration.fields.ibiTimestampEnable = target["configuration"]["ibiTimestampEnable"].value
            configuration.fields.pendingReadCapability = target["configuration"]["pendingReadCapability"].value
            tableEntry.configuration = configuration

            # Push table entry to command
            command.fields.payload[i] = tableEntry

    return command.toBytes(), I3cControllerInitBusResponse_t()

#endregion

#================================================================================#
# region I3C CONTROLLER RESET BUS SERIALIZER
#================================================================================#

def i3cControllerResetBusSerializer(id: c_uint16) -> tuple[bytes,I3cControllerResetBusResponse_t]:
    """
    This functions create and serializes an I3C CONTROLLER RESET BUS request and return the corresponding response instance.

    Arguments
    ---------
    id : c_uint16
        ID that identifies the transaction.

    Returns
    -------
    tuple
        Command serialization to be sent to the device and the response instance.

    """
    command = I3cControllerResetBusRequest_t()

    command.fields.header.id = id
    command.fields.header.code = I3cCommandCodes.I3C_CONTROLLER_RESET_BUS.value

    return command.toBytes(), I3cControllerResetBusResponse_t()

#endregion

#================================================================================#
# region I3C CONTROLLER GET TARGET DEVICES TABLE SERIALIZER
#================================================================================#

def i3cControllerGetTargetDevicesTableSerializer(id: c_uint16) -> tuple[bytes, I3cControllerGetTargetDevicesTableResponse_t]:
    """
    This functions create and serializes an I3C CONTROLLER GET TARGET DEVICES TABLE request and return the corresponding response instance.

    Arguments
    ---------
    id : c_uint16
        ID that identifies the transaction.

    Returns
    -------
    tuple
        Command serialization to be sent to the device and the response instance.

    """
    # Create command
    command = I3cControllerGetTargetDevicesTableRequest_t()

    # Set endpoint ID and command opcode
    command.fields.header.id = id
    command.fields.header.code = I3cCommandCodes.I3C_CONTROLLER_GET_TARGET_DEVICES_TABLE.value

    return command.toBytes(), I3cControllerGetTargetDevicesTableResponse_t()

#endregion

#================================================================================#
# region I3C CONTROLLER SET TARGET DEVICE CONFIGURATION SERIALIZER
#================================================================================#

def i3cControllerSetTargetDeviceConfigurationSerializer(id: c_uint16, targetAddress: c_uint8, configuration: dict) -> tuple[bytes, I3cControllerSetTargetDeviceConfigurationResponse_t]:
    """
    This functions create and serializes an I3C CONTROLLER SET TARGET DEVICE CONFIGURATION request and return the corresponding response instance.

    Arguments
    ---------
    id : c_uint16
        ID that identifies the transaction.
    targetAddress: c_uint8
            Target address whose configuration will be set.
    entry : dict
        Python dict that contains the target dynamic address and the new configuration.

    Returns
    -------
    tuple
        Command serialization to be sent to the device and the response instance.

    """
    # Create command
    command = I3cControllerSetTargetDeviceConfigurationRequest_t()

    # Set endpoint ID and command opcode

    command.fields.header.id = id
    command.fields.header.code = I3cCommandCodes.I3C_CONTROLLER_SET_TARGET_DEVICE_CONFIG.value

    # Address
    command.fields.parameters.targetAddress = targetAddress

    # Configuration
    targetConfiguration = I3cTargetConfiguration_t()
    targetConfiguration.fields.targetType = configuration["targetType"].value
    targetConfiguration.fields.acceptIbiRequest = configuration["IBIRequest"].value
    targetConfiguration.fields.acceptControllerRoleRequest = configuration["CRRequest"].value
    targetConfiguration.fields.daaUseSETDASA = configuration["daaUseSETDASA"].value
    targetConfiguration.fields.daaUseSETAASA = configuration["daaUseSETAASA"].value
    targetConfiguration.fields.daaUseENTDAA = configuration["daaUseENTDAA"].value
    targetConfiguration.fields.ibiTimestampEnable = configuration["ibiTimestampEnable"].value
    targetConfiguration.fields.pendingReadCapability = configuration["pendingReadCapability"].value
    command.fields.parameters.configuration = targetConfiguration

    return command.toBytes(), I3cControllerSetTargetDeviceConfigurationResponse_t()

#endregion

#================================================================================#
# region I3C CONTROLLER PRIVATE TRANSFER SERIALIZERS
#================================================================================#

def i3cControllerWriteSerializer(id: c_uint16, targetAddress: c_uint8, mode: TransferMode, registerAddress: list, payload: list, startWith7E: bool = True, nonStop: bool = False)  -> tuple[bytes,I3cControllerPrivateTransferResponse_t]:
    """
    This functions create and serializes an I3C CONTROLLER PRIVATE TRANSFER request and return the corresponding response instance to issue an I3C Private Write transfer.

    Arguments
    ---------
    id : c_uint16
        ID that identifies the transaction.
    targetAddress : c_uint8
        Target address.
    mode : TransferMode
        Transfer mode: I3C SDR, I3C HDR DDR, I2C, etc.
    registerAddress : list
        List that contains the target internal register address where the data is written.
    payload : list
        List that contains the data to be written.
    startWith7E: bool
        Boolean flag to indicate if the transfer must start with the broadcast address 0x7E.
    nonStop: bool
        Boolean flag to indicate whether or not the STOP condition must be issued at the end of the transfer or not.

    Returns
    -------
    tuple
        Command serialization to be sent to the device and the response instance.

    """
    command = I3cControllerPrivateTransferRequest_t()

    # Header
    command.fields.header.id = id
    command.fields.header.code = I3cCommandCodes.I3C_CONTROLLER_PRIVATE_TRANSFER.value

    # Parameters
    command.fields.parameters.targetAddress = targetAddress
    command.fields.parameters.direction = TransferDirection.WRITE.value
    command.fields.parameters.mode = mode.value
    command.fields.parameters.nonStop = c_uint8(nonStop).value
    command.fields.parameters.startWith7E = c_uint8(startWith7E).value
    command.fields.parameters.registerAddressLength = len(registerAddress)
    command.fields.parameters.payloadLength = len(payload)

    for i in range(len(registerAddress)):
        command.fields.parameters.registerAddress[i] = registerAddress[i]

    # Payload
    for i in range(len(payload)):
        command.fields.payload[i] = payload[i]

    return command.toBytes(), I3cControllerPrivateTransferResponse_t()

def i3cControllerReadSerializer(id: c_uint16, targetAddress: c_uint8, mode: TransferMode, registerAddress: list, length: c_uint16, startWith7E: bool = True, nonStop: bool = False)  -> tuple[bytes,I3cControllerPrivateTransferResponse_t]:
    """
    This functions create and serializes an I3C CONTROLLER PRIVATE TRANSFER request and return the corresponding response instance to issue an I3C Private Read transfer.

    Arguments
    ---------
    id : c_uint16
        ID that identifies the transaction.
    targetAddress : c_uint8
        Target address.
    mode : TransferMode
        Transfer mode: I3C SDR, I3C HDR DDR, I2C, etc.
    registerAddress : list
        List that contains the target internal register address where the data is written.
    length : c_uint16
        Number of bytes to read.
    startWith7E: bool
        Boolean flag to indicate if the transfer must start with the broadcast address 0x7E.
    nonStop: bool
        Boolean flag to indicate whether or not the STOP condition must be issued at the end of the transfer or not.

    Returns
    -------
    tuple
        Command serialization to be sent to the device and the response instance.

    """
    command = I3cControllerPrivateTransferRequest_t()

    # Header
    command.fields.header.id = id
    command.fields.header.code = I3cCommandCodes.I3C_CONTROLLER_PRIVATE_TRANSFER.value

    # Parameters
    command.fields.parameters.targetAddress = targetAddress
    command.fields.parameters.direction = TransferDirection.READ.value
    command.fields.parameters.mode = mode.value
    command.fields.parameters.nonStop = c_uint8(nonStop).value
    command.fields.parameters.startWith7E = c_uint8(startWith7E).value
    command.fields.parameters.registerAddressLength = len(registerAddress)
    command.fields.parameters.payloadLength = length

    for i in range(len(registerAddress)):
        command.fields.parameters.registerAddress[i] = registerAddress[i]

    return command.toBytes(), I3cControllerPrivateTransferResponse_t()

#endregion

#================================================================================#
# region I3C CONTROLLER CCC TRANSFER SERIALIZER
#================================================================================#

def i3cControllerCccTransferSerializer(id: c_uint16, targetAddress: c_uint8, direction: TransferDirection, mode: TransferMode, commandType: I3cCccType, defByte: c_uint8, ccc: c_uint8, length: c_uint16, data: list, nonStop: bool = False):
    """
    This functions create and serializes an I3C CONTROLLER CCC TRANSFER request and return the corresponding response instance.

    Arguments
    ---------
    id : c_uint16
        ID that identifies the transaction.
    targetAddress : c_uint8
        Target address.
    direction: TransferDirection
        CCC Transfer direction: Write or Read.
    mode : TransferMode
        Transfer mode: I3C SDR, I3C HDR DDR, I2C, etc.
    commandType : I3cCccType
        CCC type: With or without defining byte.
    defByte : c_uint8
        Defining byte in case cmdType indicates the command includes one
    ccc : c_uint8
        Code to send
    length : c_uint16
        Length of the data to be written or retrieved
    data : list
        List that contains the data to be written.
    nonStop: bool
        Boolean flag to indicate whether or not the STOP condition must be issued at the end of the transfer or not.

    Returns
    -------
    tuple
        Command serialization to be sent to the device and the response instance.

    """
    command = I3cControllerCccTransferRequest_t()

     # Header
    command.fields.header.id = id
    command.fields.header.code = I3cCommandCodes.I3C_CONTROLLER_CCC_TRANSFER.value

    # Parameters
    command.fields.parameters.targetAddress = targetAddress
    command.fields.parameters.direction = direction.value
    command.fields.parameters.mode = mode.value
    command.fields.parameters.nonStop = c_uint8(nonStop).value
    command.fields.parameters.type = commandType.value
    command.fields.parameters.ccc = ccc.value
    command.fields.parameters.definingByte = defByte

    if direction == TransferDirection.READ:
        command.fields.parameters.payloadLength = length
    else:
        command.fields.parameters.payloadLength = len(data)

    for i in range(len(data)):
        command.fields.payload[i] = data[i]

    return command.toBytes(), I3cControllerCccTransferResponse_t()

#endregion

#================================================================================#
# region I3C CONTROLLER TRIGGER PATTERN SERIALIZER
#================================================================================#

def i3cControllerTriggerPatternSerializer(id: c_uint16, pattern: I3cPattern) -> tuple[bytes, I3cControllerTriggerPatternResponse_t]:
    """
    This functions create and serializes an I3C CONTROLLER RIGGER PATTERN request and return the corresponding response instance.

    Arguments
    ---------
    id : c_uint16
        ID that identifies the transaction.
    pattern: I3cPattern
        Identifier of the pattern to be triggered by the Supernova.

    Returns
    -------
    tuple
        Command serialization to be sent to the device and the response instance.

    """
    command = I3cControllerTriggerPatternRequest_t()

    command.fields.header.id = id
    command.fields.header.code = I3cCommandCodes.I3C_CONTROLLER_TRIGGER_PATTERN.value
    command.fields.parameters.pattern = pattern.value

    return command.toBytes(), I3cControllerTriggerPatternResponse_t()

#endregion

#================================================================================#
# region I3C TARGET SERIALIZERS
#================================================================================#

def i3cTargetInitSerializer(id: c_uint16, memoryLayout: I3cTargetMemoryLayout_t, pid: list, bcr: c_uint8, dcr: c_uint8, staticAddress: c_uint8, mwl: c_uint16 = 1024, mrl: c_uint16 = 1024) -> tuple[bytes,I3cTargetInitResponse_t]:
    """
    This functions create and serializes an I3C TARGET INIT request and return the corresponding response instance.

    Arguments
    ---------
    id : c_uint16
            An integer number that identifies the transfer.

    memoryLayout: I3cTargetMemoryLayout_t
        Layout of the memory that the target represents.

    pid: list
        List of 6 bytes that represents the 48-bit Provisioned ID (PID).

    bcr: c_uint8
        Bus Characteristics Register.

    dcr: I3cTargetDcr_t
        Device Characteristics Register. Determines the type of device the Supernova represents.
        The enum I3cTargetDcr_t provides 3 predefined DCR values.

    staticAddress: c_uint8
        I2C static address of the Supernova.

    mwl: c_uint16
        Maximum write length as specified by SETMWL and GETMWL CCCs.

    mrl: c_uint16
        Maximum read length as specified by SETMRL and GETMRL CCCs.

    Returns
    -------
    tuple
        Command serialization to be sent to the device and the response instance.

    """
    command = I3cTargetInitRequest_t()

    command.fields.header.id = id
    command.fields.header.code = I3cCommandCodes.I3C_TARGET_INIT.value

    command.fields.parameters.memoryLayout = memoryLayout.value

    PID = I3cPID_t.from_buffer_copy(bytes(pid))

    partNo = (PID.bytes.PID_3 << 24 | PID.bytes.PID_2 << 16 | PID.bytes.PID_1 << 8 | PID.bytes.PID_0)
    randomValueFlag = (0x01 & PID.bytes.PID_4)
    vendorId = ((PID.bytes.PID_5 << 8 | PID.bytes.PID_4) >> 1)

    command.fields.parameters.partNo = partNo
    command.fields.parameters.randomValueFlag = randomValueFlag
    command.fields.parameters.vendorId = vendorId
    command.fields.parameters.bcr.byte = bcr
    command.fields.parameters.dcr = dcr.value
    command.fields.parameters.staticAddress = staticAddress
    command.fields.parameters.mwl = mwl
    command.fields.parameters.mrl = mrl

    return command.toBytes(), I3cTargetInitResponse_t()

def i3cTargetSetParametersSerializer(id: c_uint16, memoryLayout: I3cTargetMemoryLayout_t, pid: list, bcr: c_uint8, dcr: c_uint8, staticAddress: c_uint8, mwl: c_uint16 = 1024, mrl: c_uint16 = 1024) -> tuple[bytes,I3cTargetSetParametersResponse_t]:
    """
    This functions create and serializes an I3C TARGET SET PARAMETERS request and return the corresponding response instance.

    Arguments
    ---------
    id : c_uint16
            An integer number that identifies the transfer.

    memoryLayout: I3cTargetMemoryLayout_t
        Layout of the memory that the target represents.

    pid: list
        List of 6 bytes that represents the 48-bit Provisioned ID (PID).

    bcr: c_uint8
        Bus Characteristics Register.

    dcr: I3cTargetDcr_t
        Device Characteristics Register. Determines the type of device the Supernova represents.
        The enum I3cTargetDcr_t provides 3 predefined DCR values.

    staticAddress: c_uint8
        I2C static address of the Supernova.

    mwl: c_uint16
        Maximum write length as specified by SETMWL and GETMWL CCCs.

    mrl: c_uint16
        Maximum read length as specified by SETMRL and GETMRL CCCs.

    Returns
    -------
    tuple
        Command serialization to be sent to the device and the response instance.

    """
    command = I3cTargetSetParametersRequest_t()

    command.fields.header.id = id
    command.fields.header.code = I3cCommandCodes.I3C_TARGET_SET_PARAMETERS.value

    command.fields.parameters.memoryLayout = memoryLayout.value

    PID = I3cPID_t.from_buffer_copy(bytes(pid))

    partNo = (PID.bytes.PID_3 << 24 | PID.bytes.PID_2 << 16 | PID.bytes.PID_1 << 8 | PID.bytes.PID_0)
    randomValueFlag = (0x01 & PID.bytes.PID_4)
    vendorId = ((PID.bytes.PID_5 << 8 | PID.bytes.PID_4) >> 1)

    command.fields.parameters.partNo = partNo
    command.fields.parameters.randomValueFlag = randomValueFlag
    command.fields.parameters.vendorId = vendorId
    command.fields.parameters.bcr.byte = bcr
    command.fields.parameters.dcr = dcr.value
    command.fields.parameters.staticAddress = staticAddress
    command.fields.parameters.mwl = mwl
    command.fields.parameters.mrl = mrl

    return command.toBytes(), I3cTargetSetParametersResponse_t()

def i3cTargetWriteMemorySerializer(id: c_uint16, memoryAddress: c_uint16, data: list) -> tuple[bytes, I3cTargetTransferMemoryResponse_t]:
    """
    This function performs an I3C TARGET WRITE MEM command.

    Arguments
    ---------
    id : c_uint16
        ID that identifies the transaction.
    memoryAddress : c_uint16
        Register address of the target memory to start writing.
    data : list
        List that contains the data to be written.

    Returns
    -------
    tuple
        Command list with the commands to be sent to the USB device and the response instance.

    """
    command = I3cTargetTransferMemoryRequest_t()

    command.fields.header.id = id
    command.fields.header.code = I3cCommandCodes.I3C_TARGET_WRITE_MEMORY.value

    command.fields.parameters.memoryAddress = memoryAddress
    command.fields.parameters.payloadLength = len(data)

    # Data Block
    for i in range(len(data)):
        command.fields.payload[i] = data[i]

    return command.toBytes(), I3cTargetTransferMemoryResponse_t()

def i3cTargetReadMemorySerializer(id: c_uint16, memoryAddress: c_uint16, length: c_uint16) -> tuple[bytes, I3cTargetTransferMemoryResponse_t]:
    """
    This function performs an I3C TARGET READ MEM command.

    Arguments
    ---------
    id : c_uint16
        ID that identifies the transaction.
    memoryAddress : c_uint16
        Register address of the memory to start reading from.
    length : c_uint16
        Length of the data to be read.

    Returns
    -------
    tuple
        Command list with the commands to be sent to the USB device and the response instance.

    """
    command = I3cTargetTransferMemoryRequest_t()

    command.fields.header.id = id
    command.fields.header.code = I3cCommandCodes.I3C_TARGET_READ_MEMORY.value

    command.fields.parameters.memoryAddress = memoryAddress
    command.fields.parameters.payloadLength = length

    return command.toBytes(), I3cTargetTransferMemoryResponse_t()


from ..common_definitions import *

#================================================================================#
# region GPIO COMMAND DEFINITIONS
#================================================================================#

class GpioCommandCodes(Enum):
    """
    Enumeration of GPIO command codes.
    """
    GPIO_CONFIGURE_PIN              = makeCommandCode(Group.GPIO.value, CommandRole.GENERIC.value, CommandType.REQUEST_RESPONSE.value, 1)
    GPIO_DIGITAL_WRITE              = makeCommandCode(Group.GPIO.value, CommandRole.GENERIC.value, CommandType.REQUEST_RESPONSE.value, 2)
    GPIO_DIGITAL_READ               = makeCommandCode(Group.GPIO.value, CommandRole.GENERIC.value, CommandType.REQUEST_RESPONSE.value, 3)
    GPIO_SET_INTERRUPT              = makeCommandCode(Group.GPIO.value, CommandRole.GENERIC.value, CommandType.REQUEST_RESPONSE.value, 4)
    GPIO_DISABLE_INTERRUPT          = makeCommandCode(Group.GPIO.value, CommandRole.GENERIC.value, CommandType.REQUEST_RESPONSE.value, 5)
    GPIO_INTERRUPT_NOTIFICATION     = makeCommandCode(Group.GPIO.value, CommandRole.GENERIC.value, CommandType.NOTIFICATION.value, 1)

GPIO_COMMAND_NAMES = {
    result.value: result.name.replace("_", " ") for result in GpioCommandCodes
}

# endregion

#================================================================================#
# region GPIO COMMON DEFINITIONS
#================================================================================#

class GpioPinNumber(Enum):
    """
    This enum represents the possible Gpio Pin Numbers.
    """
    GPIO_1                         = 0x00
    GPIO_2                         = 0x01
    GPIO_3                         = 0x02
    GPIO_4                         = 0x03
    GPIO_5                         = 0x04
    GPIO_6                         = 0x05

class GpioLogicLevel(Enum):
    """
    This enum represents the possible logic levels for Digital I/O.
    """
    LOW                            = 0x00
    HIGH                           = 0x01

class GpioFunctionality(Enum):
    """
    This enum represents the possible functionalities for the GPIO.
    """
    DIGITAL_INPUT                  = 0x00
    DIGITAL_OUTPUT                 = 0x01

class GpioTriggerType(Enum):
    """
    This enum represents the possible Gpio interrupt trigger types.
    """
    TRIGGER_RISING_EDGE             = 0x01
    TRIGGER_FALLING_EDGE            = 0x02
    TRIGGER_BOTH_EDGES              = 0x03

# endregion

#================================================================================#
# region GPIO RESULT DEFINITIONS
#================================================================================#

class GpioResultCodes(Enum):
    """
    Enumeration of GPIO result codes.
    """
    GPIO_NOT_CONFIGURED         = makeResultCode(Group.GPIO.value, 1)
    GPIO_WRONG_CONFIGURATION    = makeResultCode(Group.GPIO.value, 2)

GPIO_RESULT_NAMES = {
    **COMMON_RESULT_NAMES,
    **{result.value: result.name for result in GpioResultCodes}
}

# endregion

#================================================================================#
# region GPIO CONFIGURE PIN
#================================================================================#

# Request ---------------------------------------------------------------------- #

class GpioConfigurePinRequestParameters_t(Structure):
    _pack_ = 1
    _fields_ = [("pinNumber", c_uint8),
                ("functionality", c_uint8)]

class GpioConfigurePinRequestFields_t(Structure):
    _pack_ = 1
    _fields_ = [("header", CommandRequestHeader_t),
                ("parameters", GpioConfigurePinRequestParameters_t)]

GpioConfigurePinRequestArray_t = c_uint8 * sizeof(GpioConfigurePinRequestFields_t)

class GpioConfigurePinRequest_t(BaseCommandRequest_t):
    _fields_ = [("data", GpioConfigurePinRequestArray_t ),
                ("fields", GpioConfigurePinRequestFields_t )]

# Response --------------------------------------------------------------------- #

class GpioConfigurePinResponseFields_t(Structure):
    _pack_ = 1
    _fields_ = [("header", CommandResponseHeader_t)]

GpioConfigurePinResponseArray_t = c_uint8 * sizeof(GpioConfigurePinResponseFields_t)

class GpioConfigurePinResponse_t(BaseCommandResponse_t):
    _fields_ = [("data", GpioConfigurePinResponseArray_t ),
                ("fields", GpioConfigurePinResponseFields_t )]

    def fromBytes(self, data):
        """
        This function set the ctypes Array data from a data buffer.
        """
        self.data = GpioConfigurePinResponseArray_t.from_buffer_copy(data)

    def toDictionary(self) -> dict:
        """
        This function returns a dictionary with the response fields.
        """
        return {
            "id": self.fields.header.id,
            "command": GPIO_COMMAND_NAMES[self.fields.header.code],
            "result": GPIO_RESULT_NAMES[self.fields.header.result]
        }

# endregion

#================================================================================#
# region GPIO DIGITAL WRITE
#================================================================================#

# Request ---------------------------------------------------------------------- #

class GpioDigitalWriteRequestParameters_t(Structure):
    _pack_ = 1
    _fields_ = [("pinNumber", c_uint8),
                ("logicLevel", c_uint8)]

class GpioDigitalWriteRequestFields_t(Structure):
    _pack_ = 1
    _fields_ = [("header", CommandRequestHeader_t),
                ("parameters", GpioDigitalWriteRequestParameters_t)]

GpioDigitalWriteRequestArray_t = c_uint8 * sizeof(GpioDigitalWriteRequestFields_t)

class GpioDigitalWriteRequest_t(BaseCommandRequest_t):
    _fields_ = [("data", GpioDigitalWriteRequestArray_t ),
                ("fields", GpioDigitalWriteRequestFields_t )]

# Response --------------------------------------------------------------------- #

class GpioDigitalWriteResponseFields_t(Structure):
    _pack_ = 1
    _fields_ = [("header", CommandResponseHeader_t)]

GpioDigitalWriteResponseArray_t = c_uint8 * sizeof(GpioDigitalWriteResponseFields_t)

class GpioDigitalWriteResponse_t(BaseCommandResponse_t):
    _fields_ = [("data", GpioDigitalWriteResponseArray_t ),
                ("fields", GpioDigitalWriteResponseFields_t )]

    def fromBytes(self, data):
        """
        This function set the ctypes Array data from a data buffer.
        """
        self.data = GpioDigitalWriteResponseArray_t.from_buffer_copy(data)

    def toDictionary(self) -> dict:
        """
        This function returns a dictionary with the response fields.
        """
        return {
            "id": self.fields.header.id,
            "command": GPIO_COMMAND_NAMES[self.fields.header.code],
            "result": GPIO_RESULT_NAMES[self.fields.header.result]
        }

# endregion

#================================================================================#
# region GPIO DIGITAL READ
#================================================================================#

# Request ---------------------------------------------------------------------- #

class GpioDigitalReadRequestParameters_t(Structure):
    _pack_ = 1
    _fields_ = [("pinNumber", c_uint8)]

class GpioDigitalReadRequestFields_t(Structure):
    _pack_ = 1
    _fields_ = [("header", CommandRequestHeader_t),
                ("parameters", GpioDigitalReadRequestParameters_t)]

GpioDigitalReadRequestArray_t = c_uint8 * sizeof(GpioDigitalReadRequestFields_t)

class GpioDigitalReadRequest_t(BaseCommandRequest_t):
    _fields_ = [("data", GpioDigitalReadRequestArray_t ),
                ("fields", GpioDigitalReadRequestFields_t )]

# Response --------------------------------------------------------------------- #

class GpioDigitalReadResponseParameters_t(Structure):
    _pack_ = 1
    _fields_ = [("logicLevel", c_uint8)]

class GpioDigitalReadResponseFields_t(Structure):
    _pack_ = 1
    _fields_ = [("header", CommandResponseHeader_t),
                ("parameters", GpioDigitalReadResponseParameters_t)]

GpioDigitalReadResponseArray_t = c_uint8 * sizeof(GpioDigitalReadResponseFields_t)

class GpioDigitalReadResponse_t(BaseCommandResponse_t):
    _fields_ = [("data", GpioDigitalReadResponseArray_t ),
                ("fields", GpioDigitalReadResponseFields_t )]

    def fromBytes(self, data):
        """
        This function set the ctypes Array data from a data buffer.
        """
        self.data = GpioDigitalReadResponseArray_t.from_buffer_copy(data)

    def toDictionary(self) -> dict:
        """
        This function returns a dictionary with the response fields.
        """
        return {
            "id": self.fields.header.id,
            "command": GPIO_COMMAND_NAMES[self.fields.header.code],
            "result": GPIO_RESULT_NAMES[self.fields.header.result],
            "logic_level": GpioLogicLevel(self.fields.parameters.logicLevel).name
        }

# endregion

#================================================================================#
# region GPIO SET INTERRUPT
#================================================================================#

# Request ---------------------------------------------------------------------- #

class GpioSetInterruptRequestParameters_t(Structure):
    _pack_ = 1
    _fields_ = [("pinNumber", c_uint8),
                ("trigger", c_uint8)]

class GpioSetInterruptRequestFields_t(Structure):
    _pack_ = 1
    _fields_ = [("header", CommandRequestHeader_t),
                ("parameters", GpioSetInterruptRequestParameters_t)]

GpioSetInterruptRequestArray_t = c_uint8 * sizeof(GpioSetInterruptRequestFields_t)

class GpioSetInterruptRequest_t(BaseCommandRequest_t):
    _fields_ = [("data", GpioSetInterruptRequestArray_t ),
                ("fields", GpioSetInterruptRequestFields_t )]

# Response --------------------------------------------------------------------- #

class GpioSetInterruptResponseFields_t(Structure):
    _pack_ = 1
    _fields_ = [("header", CommandResponseHeader_t)]

GpioSetInterruptResponseArray_t = c_uint8 * sizeof(GpioSetInterruptResponseFields_t)

class GpioSetInterruptResponse_t(BaseCommandResponse_t):
    _fields_ = [("data", GpioSetInterruptResponseArray_t ),
                ("fields", GpioSetInterruptResponseFields_t )]

    def fromBytes(self, data):
        """
        This function set the ctypes Array data from a data buffer.
        """
        self.data = GpioSetInterruptResponseArray_t.from_buffer_copy(data)


    def toDictionary(self) -> dict:
        """
        This function returns a dictionary with the response fields.
        """
        return {
            "id": self.fields.header.id,
            "command": GPIO_COMMAND_NAMES[self.fields.header.code],
            "result": GPIO_RESULT_NAMES[self.fields.header.result]
        }

# endregion

#================================================================================#
# region GPIO DISABLE INTERRUPT
#================================================================================#

# Request ---------------------------------------------------------------------- #

class GpioDisableInterruptRequestParameters_t(Structure):
    _pack_ = 1
    _fields_ = [("pinNumber", c_uint8)]

class GpioDisableInterruptRequestFields_t(Structure):
    _pack_ = 1
    _fields_ = [("header", CommandRequestHeader_t),
                ("parameters", GpioDisableInterruptRequestParameters_t)]

GpioDisableInterruptRequestArray_t = c_uint8 * sizeof(GpioDisableInterruptRequestFields_t)

class GpioDisableInterruptRequest_t(BaseCommandRequest_t):
    _fields_ = [("data", GpioDisableInterruptRequestArray_t ),
                ("fields", GpioDisableInterruptRequestFields_t )]

# Response --------------------------------------------------------------------- #

class GpioDisableInterruptResponseFields_t(Structure):
    _pack_ = 1
    _fields_ = [("header", CommandResponseHeader_t)]

GpioDisableInterruptResponseArray_t = c_uint8 * sizeof(GpioDisableInterruptResponseFields_t)

class GpioDisableInterruptResponse_t(BaseCommandResponse_t):
    _fields_ = [("data", GpioDisableInterruptResponseArray_t ),
                ("fields", GpioDisableInterruptResponseFields_t )]

    def fromBytes(self, data):
        """
        This function set the ctypes Array data from a data buffer.
        """
        self.data = GpioDisableInterruptResponseArray_t.from_buffer_copy(data)

    def toDictionary(self) -> dict:
        """
        This function returns a dictionary with the response fields.
        """
        return {
            "id": self.fields.header.id,
            "command": GPIO_COMMAND_NAMES[self.fields.header.code],
            "result": GPIO_RESULT_NAMES[self.fields.header.result]
        }

# endregion

#================================================================================#
# region GPIO NOTIFICATION
#================================================================================#

class GpioInterruptNotificationParameters_t(Structure):
    _pack_ = 1
    _fields_ = [("pinNumber", c_uint8)]

class GpioInterruptNotificationFields_t(Structure):
    _pack_ = 1
    _fields_ = [("header", CommandResponseHeader_t),
                ("parameters", GpioInterruptNotificationParameters_t)]

gpioInterruptNotificationArray_t = c_uint8 * sizeof(GpioInterruptNotificationFields_t)

class GpioInterruptNotification_t(Union):
    _fields_ = [("data", gpioInterruptNotificationArray_t ),
                ("fields", GpioInterruptNotificationFields_t )]

    def toDictionary(self) -> dict:
        return {
            "id": self.fields.header.id,
            "command": GPIO_COMMAND_NAMES[self.fields.header.code],
            "result": GPIO_RESULT_NAMES[self.fields.header.result],
            "pin_number": GpioPinNumber(self.fields.parameters.pinNumber).name
        }

    def __str__(self) -> str:
        return str(self.toDictionary())

# endregion